;; Generated package description from devicetree-ts-mode.el  -*- no-byte-compile: t -*-
(define-package "devicetree-ts-mode" "0.3.0.20240117.132538" "Tree-sitter support for DTS" '((emacs "29.1")) :commit "3a4819f76c2af3437b95d5cee46b7e0252a3a8f3" :authors '(("Aleksandr Vityazev" . "avityazev@disroot.org")) :maintainer '("Devicetree-ts Development" . "~akagi/devicetree-ts-mode@lists.sr.ht") :keywords '("languages" "devicetree" "tree-sitter") :url "https://sr.ht/~akagi/devicetree-ts-mode")
